The package {{ package }} build has experienced an issue.
Please know that the explanation was provided by AI and may be incorrect.
{% if report_certainty %}
{% if certainty > 0 %}
    In this case, we are {{ "%.2f" | format(certainty) }}% certain of the response {{ emoji_face }}.
{% endif %}
{% endif %}

<details open>
    <summary>Description</summary>
    {{ explanation }}
</details>
<details>
{#
    Formatted so that we don't trigger GitLab markdown
#}
<summary>Snippets</summary>
<ul>
{% for snippet in snippets -%}
<li><div><b>Line {{ snippet.line_number }}:</b> <code>{{ snippet.text | e }}</code><br>{{ snippet.explanation.text | e }}</div></li>
{%- endfor %}
</ul>
</details>
<details>
    <summary>Logs</summary>
    <p>Log Detective analyzed the following logs files to provide an explanation:</p>
    <ul>
        <li>
            <a href="{{ log_url }}">{{ log_url }}</a>
        </li>
    </ul>
    <p>
        Additional logs are available from:
        <ul>
            <li>
                <a href="{{ artifacts_url }}">artifacts.zip</a>
            </li>
        </ul>
    </p>
    <p>
        Please know that these log files are automatically removed after some
        time, so you might need a backup.
    </p>
</details>
<details>
    <summary>Help</summary>
    <p>Don't hesitate to reach out.</p>
    <ul>
        <li>
            <a href="https://github.com/fedora-copr/logdetective">Upstream</a>
        </li>
        <li>
            <a href="https://github.com/fedora-copr/logdetective/issues">Issue tracker</a>
        </li>
        <li>
            <a href="https://redhat.enterprise.slack.com/archives/C06DWNVKKDE">Slack</a>
        </li>
        <li>
            <a href="https://log-detective.com/documentation">Documentation</a>
        </li>
    </ul>
</details>

<hr>

This explanation was provided by AI (<a href="https://logdetective.com">Log Detective</a>).
Always review AI generated content prior to use.
Was the provided feedback accurate and helpful?
<br>
Please vote with :thumbsup:
or :thumbsdown: to help us improve.
<br>
<i>If this Log Detective report contains harmful content,
please use the <a href="https://docs.gitlab.com/user/report_abuse/">Gitlab reporting feature for harmful content</a>
and contact the <a href="https://github.com/fedora-copr/logdetective/issues">Log Detective developers</a>.</i>
